﻿using System;
using System.Collections.Generic;
using UnityEngine;
using Verse;
using Verse.AI;
using Verse.Sound;
using RimWorld;

namespace RecreationalSlotMachine
{
	public enum MoodBoostStackingMode { Concurrent, Extend, Refresh }
	public class RecSlotMachineSettings : ModSettings
	{
		public bool winsGiveMoodBoost = true;
		public bool winsGiveGold = true;
		public float goldSlotChance = 0.25f;
		public int payoutMin = 1;
		public int payoutMax = 1;
		public float moodBoostAmount = 5f;
		public float moodBoostDurationDays = 2f;
		public bool[] radiosMoodBoostStackingMode = new bool[] { true, false, false };
		public bool[] radiosPriorMoodBoostStackingMode = new bool[] { false, false, false };
		public MoodBoostStackingMode moodBoostStackingMode = MoodBoostStackingMode.Concurrent;
		public int moodBoostStackAmount = 1;

		public string moodBoostStackAmountBuffer = "1";
		public string payoutMinBuffer = "1";
		public string payoutMaxBuffer = "1";
		public string moodBoostAmountBuffer = "5";
		public string moodBoostDurationDaysBuffer = "2";
		public Vector2 scrollPosition = new Vector2(1f, 1f);
		public float gameLengthSeconds = 2f;
		public string gameLengthSecondsBuffer = "2";
		public int gameLengthTicks = 300;

		// Gives radio button logic to a list of bools tied to a set of checkboxes. Not used in this mod but was going to be and probably will come in handy later.
		public void RadioSettingsController(bool[] radioButtonsBoolsCurrently, bool[] radioButtonsBoolsPreviously)
		{
			bool allRadiosDeactive = true;
			for(int i = 0; i < radioButtonsBoolsCurrently.Length; i++)
            {
				allRadiosDeactive = !radioButtonsBoolsCurrently[i] & allRadiosDeactive;
            }
			if (allRadiosDeactive)
			{
				radioButtonsBoolsCurrently[0] = true;
				radioButtonsBoolsPreviously[0] = false;
			}
			for (int i = 0; i < radioButtonsBoolsCurrently.Length; i++)
            {
				if (radioButtonsBoolsCurrently[i] & !radioButtonsBoolsPreviously[i])
                {
					radioButtonsBoolsPreviously[i] = true;
					for (int j = 0; j < radioButtonsBoolsCurrently.Length; j++)
                    {
						if (i == j)
							continue;
						radioButtonsBoolsCurrently[j] = false;
						radioButtonsBoolsPreviously[j] = false;
                    }
					break;
                }
            }
		}
		public void UpdateMoodBoostStackingMode()
        {
			if (radiosMoodBoostStackingMode[0])
				moodBoostStackingMode = MoodBoostStackingMode.Concurrent;
			else if (radiosMoodBoostStackingMode[1])
				moodBoostStackingMode = MoodBoostStackingMode.Extend;
			else if (radiosMoodBoostStackingMode[2])
				moodBoostStackingMode = MoodBoostStackingMode.Refresh;
			else
				moodBoostStackingMode = MoodBoostStackingMode.Concurrent;
        }
		public void UpdateGameLengthTicks()
        {
			gameLengthTicks = (int)(gameLengthSeconds * 60f);
		}
		public void CorrectUserInput()
		{
			int temp = payoutMax;
			payoutMax = Math.Max(payoutMin, payoutMax);
			payoutMin = Math.Min(payoutMin, temp);
			payoutMaxBuffer = payoutMax.ToString();
			payoutMinBuffer = payoutMin.ToString();
			if (moodBoostStackAmount < 1)
            {
				moodBoostStackAmount = 1;
				moodBoostStackAmountBuffer = "1";
            }
		}
		public override void ExposeData()
		{
            Scribe_Values.Look(ref winsGiveMoodBoost, "winsGiveMoodBoost", true);
			Scribe_Values.Look(ref winsGiveGold, "winsGiveGold", true);
			Scribe_Values.Look(ref goldSlotChance, "goldSlotChance", 0.25f);
			Scribe_Values.Look(ref payoutMin, "payoutMin", 1);
			Scribe_Values.Look(ref payoutMax, "payoutMax", 1);
			Scribe_Values.Look(ref moodBoostAmount, "moodBoostAmount", 5f);
			Scribe_Values.Look(ref moodBoostDurationDays, "moodBoostDurationDays", 2f);
			Scribe_Values.Look(ref gameLengthSeconds, "gameLengthSeconds", 2f);
			Scribe_Values.Look(ref moodBoostStackAmount, "moodBoostStackAmount", 1);
			UpdateGameLengthTicks();
			payoutMinBuffer = payoutMin.ToString();
			payoutMaxBuffer = payoutMax.ToString();
			moodBoostAmountBuffer = moodBoostAmount.ToString();
			moodBoostDurationDaysBuffer = moodBoostDurationDays.ToString();
			moodBoostStackAmountBuffer = moodBoostStackAmount.ToString();
			gameLengthSecondsBuffer = gameLengthSeconds.ToString();
			base.ExposeData();
		}
	}

	[StaticConstructorOnStartup]
	public class RecreationalSlotMachine : Mod
	{
		public static RecSlotMachineSettings settings;
		public static ThingDef[] reelFrameThingDefs;
		public static Thing[] reelFrameThings;
		public static Mesh[] reelMeshes;
		public static Material[] reelMaterials;
		public static int blurFrameCount;
		public static bool initialised = false;
		public RecreationalSlotMachine(ModContentPack content) : base(content)
		{
			settings = GetSettings<RecSlotMachineSettings>();
		}
		public override void DoSettingsWindowContents(Rect inRect)
		{
			Listing_Standard listingStandard = new Listing_Standard();
            listingStandard.Begin(inRect);
            string winsGiveGoldText = "Winning gives gold";
			string winsGiveGoldToolTip = "Gold falls out of the machine. Who put that in there?";
			string winsGiveMoodBoostText = "Winning gives mood boost";
			string winsGiveMoodBoostToolTip = "Winning feels good";
			listingStandard.Label("Chance for 1 slot to land on gold: " + (settings.goldSlotChance * 100f).ToString() + "%");
			settings.goldSlotChance = listingStandard.Slider(settings.goldSlotChance, 0.012f, 1f);
			listingStandard.Label("Length of 1 game in seconds. The pawn gets 20 seconds to play the machine at a time.");
			listingStandard.TextFieldNumeric(ref settings.gameLengthSeconds, ref settings.gameLengthSecondsBuffer, 0.1f);
			float winChanceNum = Mathf.Pow(settings.goldSlotChance, 3f) * 100f;
			int gamesCompletable = (int)(20.017f / settings.gameLengthSeconds);
			float overallWinChanceNum = winChanceNum * gamesCompletable;
			string winChance = winChanceNum.ToString();
			if (winChance.Length > 5)
				winChance = winChance.Substring(0, 5);
			string overallWinChance = overallWinChanceNum.ToString();
			if (overallWinChance.Length > 5)
				overallWinChance = overallWinChance.Substring(0, 5);
			listingStandard.Label("Pawns can complete " + gamesCompletable.ToString() + " games per 20-second session. Pawns have a " + winChance + "% chance per game (" + overallWinChance + "% overall) to win");
			listingStandard.Label("");
			listingStandard.CheckboxLabeled(winsGiveGoldText, ref settings.winsGiveGold, winsGiveGoldToolTip);
			listingStandard.Label("Minimum payout");
			listingStandard.IntEntry(ref settings.payoutMin, ref settings.payoutMinBuffer);
			listingStandard.Label("Maximum payout");
			listingStandard.IntEntry(ref settings.payoutMax, ref settings.payoutMaxBuffer);
			listingStandard.Label("");
			listingStandard.CheckboxLabeled(winsGiveMoodBoostText, ref settings.winsGiveMoodBoost, winsGiveMoodBoostToolTip);
			listingStandard.Label("Mood boost stack limit");
			listingStandard.IntEntry(ref settings.moodBoostStackAmount, ref settings.moodBoostStackAmountBuffer);
			listingStandard.Label("Amount of mood per boost");
			listingStandard.TextFieldNumeric(ref settings.moodBoostAmount, ref settings.moodBoostAmountBuffer);
			listingStandard.Label("How many days the mood boost will last");
			listingStandard.TextFieldNumeric(ref settings.moodBoostDurationDays, ref settings.moodBoostDurationDaysBuffer);
			settings.UpdateGameLengthTicks();
			settings.CorrectUserInput();
            listingStandard.End();
            base.DoSettingsWindowContents(inRect);
		}
		public override string SettingsCategory()
		{
			return "RecreationalSlotMachine";
		}
		public static void InitialiseTextures()
        {
			if (!initialised)
			{
				reelFrameThingDefs = new ThingDef[]
				{
					MyDefs.Mote_ReelOutcomeA,
					MyDefs.Mote_ReelOutcomeB,
					MyDefs.Mote_ReelBlurA,
					MyDefs.Mote_ReelBlurB,
					MyDefs.Mote_ReelBlurC,
					MyDefs.Mote_ReelBlurD,
					MyDefs.Mote_ReelBlurE,
					MyDefs.Mote_ReelBlurF,
					MyDefs.Mote_ReelBlurG,
					MyDefs.Mote_ReelBlurH,
					MyDefs.Mote_ReelBlurI,
					MyDefs.Mote_ReelBlurJ
				};
				blurFrameCount = reelFrameThingDefs.Length - 2;
				reelFrameThings = new Thing[reelFrameThingDefs.Length];
				reelMeshes = new Mesh[reelFrameThingDefs.Length];
				reelMaterials = new Material[reelFrameThingDefs.Length];
				for (int i = 0; i < reelFrameThings.Length; i++)
				{
					reelFrameThings[i] = ThingMaker.MakeThing(reelFrameThingDefs[i]);
					reelMeshes[i] = reelFrameThings[i].Graphic.MeshAt(Rot4.North);
					reelMaterials[i] = reelFrameThings[i].Graphic.MatSingle;
				}
				initialised = true;
			}
        }
	}
	public class SlotMachineComp : ThingComp
    {
		public bool initialised = false;
		public SlotMachineGameEventManager eventManager = new SlotMachineGameEventManager();
		public bool justRespawned = false;
        public override void PostSpawnSetup(bool respawningAfterLoad)
        {
            base.PostSpawnSetup(respawningAfterLoad);
			justRespawned = respawningAfterLoad;
			for (int i = 0; i < 3; i++)
				eventManager.reels[i] = new SlotReel(i);
			Vector3 middle = parent.Position.ToVector3() + new Vector3(0.5045f, 0f, 0.5f);
			Vector3 offset = new Vector3(0.188f, 0f, 0f);
			middle.z += 0.445f;
			eventManager.reels[0].drawLocation = middle - offset;
			eventManager.reels[1].drawLocation = middle;
			eventManager.reels[2].drawLocation = middle + offset;
			eventManager.reels[0].drawLocation.y = 5.5f;
			eventManager.reels[1].drawLocation.y = 5.5f;
			eventManager.reels[2].drawLocation.y = 5.5f;
			eventManager.gameStarted = false;
			initialised = true;
		}
        public override void PostDeSpawn(Map map)
		{
			initialised = false;
			base.PostDeSpawn(map);
        }
        public override void CompTick()
        {
            base.CompTick();
			RecreationalSlotMachine.InitialiseTextures(); // Only happens once. This is the only safe place I know of right now to call it from.
			foreach(SlotReel reel in eventManager.reels)
            {
				if (reel.state == SlotReelState.Spinning)
				{
					if (reel.ticksRemainingForHardCutoff > 0)
						reel.ticksRemainingForHardCutoff--;
					if (reel.ticksRemainingForHardCutoff == 0)
					{
						reel.state = SlotReelState.Loss;
					}
					reel.UpdateDrawState();
				}
            }
        }
        public override void PostDraw()
        {
            base.PostDraw();
			if (initialised)
			{
				Graphics.DrawMesh(eventManager.reels[0].drawingMesh, eventManager.reels[0].drawLocation, Rot4.North.AsQuat, eventManager.reels[0].drawingMaterial, 0);
				Graphics.DrawMesh(eventManager.reels[1].drawingMesh, eventManager.reels[1].drawLocation, Rot4.North.AsQuat, eventManager.reels[1].drawingMaterial, 0);
				Graphics.DrawMesh(eventManager.reels[2].drawingMesh, eventManager.reels[2].drawLocation, Rot4.North.AsQuat, eventManager.reels[2].drawingMaterial, 0);
			}
			justRespawned = false;
		}
	}
	public enum SlotReelState { Loss, Win, Spinning }
	public class SlotReel
	{
		public SlotReelState state = SlotReelState.Loss;
		public Vector3 drawLocation;
		public int ticksRemainingForHardCutoff; // Hard cutoff occurs when a pawn leaves an unfinished game
		public int id;
		public Mesh drawingMesh;
		public Material drawingMaterial;
		public SlotReel(int id)
        {
			state = SlotReelState.Spinning;
			this.id = id;
			ticksRemainingForHardCutoff = 1;
        }
		public void UpdateDrawState()
        {
			switch(state)
			{
				case SlotReelState.Loss:
					drawingMaterial = RecreationalSlotMachine.reelMaterials[0];
					drawingMesh = RecreationalSlotMachine.reelMeshes[0];
					break;
				case SlotReelState.Win:
					drawingMaterial = RecreationalSlotMachine.reelMaterials[1];
					drawingMesh = RecreationalSlotMachine.reelMeshes[1];
					break;
				case SlotReelState.Spinning:
					int thisFrame = Rand.Range(2, 2 + RecreationalSlotMachine.blurFrameCount);
					drawingMaterial = RecreationalSlotMachine.reelMaterials[thisFrame];
					drawingMesh = RecreationalSlotMachine.reelMeshes[thisFrame];
					break;
			}
        }
    }
	public class SlotMachineGameEventManager
    {
		public List<SlotMachineGameEvent> gameEvents = new List<SlotMachineGameEvent>();
		public SlotReel[] reels = new SlotReel[3];
		public bool gameStarted = false;
		public bool gameWon = false;
		public int goldReelCount;
		public bool NextEvent(int tickNow)
        {
			if (gameEvents.Count > 0)
            {
				if(gameEvents[0].AttemptFireEvent(tickNow))
                {
					if (gameEvents[0].eventType == SlotEventType.Start)
					{
						goldReelCount = 0;
						gameWon = false;
					}
					else if ((gameEvents[0].eventType == SlotEventType.Stop) & (gameEvents[0].reel.state == SlotReelState.Win))
                    {
						goldReelCount++;
                    }
					return true;
                }
            }
			return false;
        }
		public void CalculateNewEvents(int ticksLength, float gameTime)
        {
			if (gameTime < 0.1f) gameTime = 0.1f;
			gameEvents.Clear();
			float startInterval = gameTime * 0.95f;
			float reel1Interval = gameTime * 0.55f;
			float reel2Interval = gameTime * 0.35f;
			float reel3Interval = gameTime * 0.15f;
			float sessionLength = ticksLength / 60;
			while (sessionLength >= gameTime)
            {
				sessionLength -= gameTime;
				int thisStart = (int)((sessionLength + startInterval) * 60f);
				int thisReel1 = (int)((sessionLength + reel1Interval) * 60f);
				int thisReel2 = (int)((sessionLength + reel2Interval) * 60f);
				int thisReel3 = (int)((sessionLength + reel3Interval) * 60f);
				gameEvents.Add(new SlotMachineGameEvent(thisStart, reels[0], SlotEventType.Start));
				gameEvents.Add(new SlotMachineGameEvent(thisStart, reels[1], SlotEventType.Start));
				gameEvents.Add(new SlotMachineGameEvent(thisStart, reels[2], SlotEventType.Start));
				gameEvents.Add(new SlotMachineGameEvent(thisReel1, reels[0], SlotEventType.Stop));
				gameEvents.Add(new SlotMachineGameEvent(thisReel2, reels[1], SlotEventType.Stop));
				gameEvents.Add(new SlotMachineGameEvent(thisReel3, reels[2], SlotEventType.Stop));
			}
			gameStarted = true;
        }
    }
	public enum SlotEventType { Start, Stop }
	public class SlotMachineGameEvent
    {
		public int tickHappeningOn;
		public bool eventFiredYet = false;
		public SlotEventType eventType;
		public SlotReel reel;
		public SlotMachineGameEvent(int tickHappeningOn, SlotReel reel, SlotEventType type)
        {
			if (tickHappeningOn < 0) tickHappeningOn = 0;
			this.tickHappeningOn = tickHappeningOn;
			eventType = type;
			this.reel = reel;
        }
		public bool AttemptFireEvent(int tickNow)
        {
			if (tickNow <= tickHappeningOn)
			{
				if (eventType == SlotEventType.Start)
                {
					reel.state = SlotReelState.Spinning;
					reel.ticksRemainingForHardCutoff = 1205;
                }
				else
					reel.state = (Rand.Range(0f, 1f) <= RecreationalSlotMachine.settings.goldSlotChance) ? SlotReelState.Win : SlotReelState.Loss;
				reel.UpdateDrawState();
				return true;
			}
			return false;
        }
    }

	public class JobDriver_PlaySlotMachine : JobDriver
	{
		public override bool TryMakePreToilReservations(bool errorOnFailed)
		{
			return this.pawn.Reserve(this.job.targetA, this.job, this.job.def.joyMaxParticipants, 0, null, errorOnFailed);
		}
		public int toilDuration = 1205;

        protected override IEnumerable<Toil> MakeNewToils()
		{
			this.EndOnDespawnedOrNull(TargetIndex.A, JobCondition.Incompletable);
			yield return Toils_Goto.GotoCell(TargetThingA.InteractionCell, PathEndMode.OnCell);
			Toil toil = new Toil();
			toil.initAction = delegate ()
			{
				this.job.locomotionUrgency = LocomotionUrgency.Walk;
				TargetThingA.TryGetComp<SlotMachineComp>().eventManager.gameStarted = false;
			};
			toil.tickAction = delegate ()
			{
				SlotMachineComp comp = TargetThingA.TryGetComp<SlotMachineComp>();
				if (comp.initialised)
				{
					if (comp.justRespawned)
                    {
						comp.justRespawned = false;
						base.EndJobWith(JobCondition.Succeeded);
						return;
					}
					if (!comp.eventManager.gameStarted)
					{
						comp.eventManager.CalculateNewEvents(toilDuration, RecreationalSlotMachine.settings.gameLengthSeconds);
					}
					bool playStart = false;
					while (comp.eventManager.NextEvent(ticksLeftThisToil))
					{
						if (comp.eventManager.gameEvents[0].eventType == SlotEventType.Start)
							playStart = true;
						else
						{
							MyDefs.MapOnlyDragSlider.PlayOneShot(new TargetInfo(pawn.Position, pawn.Map));
						}
						if (comp.eventManager.goldReelCount == 3 & !comp.eventManager.gameWon)
						{
							comp.eventManager.gameWon = true;
							MyDefs.MapOnlyTinyBell.PlayOneShot(new TargetInfo(pawn.Position, pawn.Map));
							if (RecreationalSlotMachine.settings.winsGiveMoodBoost)
							{
								MyDefs.WonSlotMachineGame.durationDays = RecreationalSlotMachine.settings.moodBoostDurationDays;
								MyDefs.WonSlotMachineGame.stackLimit = RecreationalSlotMachine.settings.moodBoostStackAmount;
								pawn.needs.mood.thoughts.memories.TryGainMemory(MyDefs.WonSlotMachineGame);
							}
							if (RecreationalSlotMachine.settings.winsGiveGold)
							{
								int goldRewarded = Rand.RangeInclusive(RecreationalSlotMachine.settings.payoutMin, RecreationalSlotMachine.settings.payoutMax);
								if (goldRewarded > 0)
								{
									Thing gold = ThingMaker.MakeThing(ThingDefOf.Gold);
									gold.stackCount = goldRewarded;
									GenSpawn.Spawn(gold, pawn.Position, Map);
								}
							}
						}
						comp.eventManager.gameEvents.RemoveAt(0);
					}
					if (playStart)
					{
						SoundDefOf.FlickSwitch.PlayOneShot(new TargetInfo(pawn.Position, pawn.Map));
					}
				}
				if (Find.TickManager.TicksGame > this.startTick + this.job.def.joyDuration)
				{
					base.EndJobWith(JobCondition.Succeeded);
					return;
				}
				JoyUtility.JoyTickCheckEnd(this.pawn, JoyTickFullJoyAction.EndJob, 1f, (Building)base.TargetThingA);
			};
			
			toil.socialMode = RandomSocialMode.SuperActive;
			toil.defaultCompleteMode = ToilCompleteMode.Delay;
			toil.defaultDuration = toilDuration;
			toil.AddFinishAction(delegate
			{
				JoyUtility.TryGainRecRoomThought(this.pawn);
			});
			yield return toil;
			yield break;
		}

		public override object[] TaleParameters()
		{
			return new object[]
			{
				this.pawn,
				base.TargetA.Thing.def
			};
		}
	}
	[DefOf]
	public class MyDefs
	{
		static MyDefs()
		{
			DefOfHelper.EnsureInitializedInCtor(typeof(MyDefs));
		}

		public static ThoughtDef WonSlotMachineGame;
		public static SoundDef MapOnlyDragSlider;
		public static SoundDef MapOnlyTinyBell;
		public static ThingDef SlotMachine;
		public static ThingDef Mote_ReelBlurA;
		public static ThingDef Mote_ReelBlurB;
		public static ThingDef Mote_ReelBlurC;
		public static ThingDef Mote_ReelBlurD;
		public static ThingDef Mote_ReelBlurE;
		public static ThingDef Mote_ReelBlurF;
		public static ThingDef Mote_ReelBlurG;
		public static ThingDef Mote_ReelBlurH;
		public static ThingDef Mote_ReelBlurI;
		public static ThingDef Mote_ReelBlurJ;
		public static ThingDef Mote_ReelOutcomeA;
		public static ThingDef Mote_ReelOutcomeB;
	}
}